from typing import List, Tuple

from Bio.pairwise2 import align
from Bio.SeqIO import parse
from numpy.random import random
from tensorflow import Tensor, int64
from tensorflow.data import TFRecordDataset
from tensorflow.io import TFRecordWriter, VarLenFeature, parse_single_example
from tensorflow.train import Example, Feature, Features, Int64List

from constants import *


def align_sequences(sequence, label) -> Tuple[str, str]:
    """
    Align the altered sequence with the reference sequence to obtain a same length output
    """
    alignments = align.globalxx(label, sequence)
    best_alignment = alignments[0]
    aligned_seq, aligned_label, _, _, _ = best_alignment
    return aligned_seq, aligned_label


def generate_example(sequence, label) -> bytes:
    """
    Create a binary-string for each sequence containing the sequence and the bases' counts
    """
    aligned_seq, aligned_label = align_sequences(sequence, label)
    schema = {
        "sequence": Feature(int64_list=Int64List(value=encode_sequence(aligned_seq))),
        "label": Feature(int64_list=Int64List(value=encode_sequence(aligned_label))),
    }
    example = Example(features=Features(feature=schema))
    return example.SerializeToString()


def encode_sequence(sequence) -> List[int]:
    """
    Encode the DNA sequence using the indices of the BASES constant
    """
    encoded_sequence = [BASES.index(element) for element in sequence]
    return encoded_sequence


def read_fastq(data_file, label_file) -> List[bytes]:
    """
    Parses a data and a label FASTQ files and generates a List of serialized Examples
    """
    examples = []
    with open(data_file) as data, open(label_file) as labels:
        for element, label in zip(parse(data, "fastq"), parse(labels, "fastq")):
            example = generate_example(sequence=str(element.seq), label=str(label.seq))
            examples.append(example)
    return examples


def create_dataset(data_file, label_file, dataset_split=[0.8, 0.1, 0.1]) -> None:
    """
    Create a training, evaluation and test dataset with a 80/10/10 split respectively
    """
    data = read_fastq(data_file, label_file)
    with TFRecordWriter(TRAIN_DATASET) as training, TFRecordWriter(
        TEST_DATASET
    ) as test, TFRecordWriter(EVAL_DATASET) as evaluation:
        for element in data:
            if random() < dataset_split[0]:
                training.write(element)
            elif random() < dataset_split[0] + dataset_split[1]:
                evaluation.write(element)
            else:
                test.write(element)


def process_input(byte_string) -> Tuple[Tensor, Tensor]:
    """
    Parse a byte-string into an Example object
    """
    schema = {
        "sequence": VarLenFeature(dtype=int64),
        "label": VarLenFeature(dtype=int64),
    }
    features = parse_single_example(byte_string, features=schema)
    return features["sequence"], features["label"]


def read_dataset(filepath) -> TFRecordDataset:
    """
    Read TFRecords files and generate a dataset
    """
    data_input = TFRecordDataset(filenames=filepath)
    dataset = data_input.map(map_func=process_input)
    shuffled_dataset = dataset.shuffle(buffer_size=10000, seed=42)
    batched_dataset = shuffled_dataset.batch(batch_size=BATCH_SIZE).repeat(count=EPOCHS)
    return batched_dataset


def dataset_creation(
    data_file, label_file
) -> Tuple[TFRecordDataset, TFRecordDataset, TFRecordDataset]:
    """
    Generate the TFRecord files and split them into training, validation and test data
    """
    create_dataset(data_file, label_file)
    train_data = read_dataset(TRAIN_DATASET)
    eval_data = read_dataset(EVAL_DATASET)
    test_data = read_dataset(TEST_DATASET)
    return train_data, eval_data, test_data
